/****
 * 2013/04/17
 * Version 1.0
 * 
 Testing code for the motor controlling part on HCR
****/

#include <Servo.h>
#include <Metro.h>
#include <Wire.h>

Metro DataTrans = Metro(200,true);
Metro BehaviorInterval = Metro(25,true);

#define LF 0
#define RT 1
#define Pbn  8

int Lduration,Rduration;
boolean LcoderDir,RcoderDir;
const byte encoder0pinA = 2;
const byte encoder0pinB = 4;
byte encoder0PinALast;
const byte encoder1pinA = 3;
const byte encoder1pinB = 5;
byte encoder1PinALast;
int RotationCoder[2];
int pastCoder[2];

long totalCoder[2];

int _speedtarget[2];
double _lasterror[2];
double _preverror[2];

float _proportion,_integral,_derivative,_maximum,_minimum;

Servo MotorLeft;
Servo MotorRight;
float _speedleft,_speedright;
float _perimeterA,_FirmPulsePG;

byte wheelDir = 0x00;
int _CMDspeed[2] = {
  0,0};

#define LMotorDir 10
#define RMotorDir 9

void setup()
{
  delay(1000);
//  Wire.begin(10);
//  Wire.onRequest(requestEvent); // register event
//  Wire.onReceive(receiveEvent); // register event

  MotorLeft.attach(LMotorDir,1000,2000);
  MotorRight.attach(RMotorDir,1000,2000);

  _speedtarget[LF] = 0;
  _speedtarget[RT] = 0;

  MotorLeft.write(90);
  MotorRight.write(90);
  
//  while(true);  

  Serial.begin(57600);
  CoderInit();

  _perimeterA = 43.6*1000;
  _FirmPulsePG = 4250;

  _proportion = 3;//4
  _integral = 0.5;//1
  _derivative = 0.6;//1
  _maximum = 500;
  _minimum = _maximum*(-1);

  //  Serial.print("Perimeter\t");
  //  Serial.println(_perimeterA);
  //  while(!Serial.available());
}

void loop()
{
  if(DataTrans.check())
  {    
//    Serial.print(_speedtarget[LF]);
//    Serial.print(",");
//    Serial.print(_speedtarget[RT]);
//    Serial.print(",");
//    Serial.print(pastCoder[LF]);
//    Serial.print(",");
//    Serial.print(pastCoder[RT]);
//    Serial.print("\n");

    /*
    Serial.print(Lduration);
    Serial.print(",");
    Serial.println(Rduration);
     */
  }
  if(BehaviorInterval.check())
  {
    static int lastLspeed = 0;
    static int lastRspeed = 0;
    ResentSpeed();

//    Motor(2000,LF);
//    Motor(1000,RT);

    int _Loutput,_Routput;

    float Lpara,Rpara;
    Lpara = TVPIDcal(_speedleft,true);
    Rpara = TVPIDcal(_speedright,false);

    //    Serial.print(Lpara);

    _Loutput = int(TVAffect(Lpara));
    _Routput = int(TVAffect(Rpara));

    Motor(_Loutput,LF);
    Motor(_Routput,RT);

    //    lastLspeed = _Loutput;
    //    lastRspeed = _Routput;

    //    Serial.print("LW:");
    //    Serial.print(_speedleft);
    //    Serial.print(",");
    //    Serial.print(_speedtarget[LF]);
    //    Serial.print(",");
    //    Serial.print(Lpara);
    //    Serial.print(",");
    //    Serial.print(_Loutput);
    //    Serial.print("\tRW:");
    //    Serial.print(_speedright);
    //    Serial.print(",");
    //    Serial.print(_speedtarget[RT]);
    //    Serial.print(",");
    //    Serial.print(Rpara);
    //    Serial.print(",");
    //    Serial.println(_Routput);
  }
}


void CoderInit()
{
  LcoderDir = true;
  RcoderDir = true;	

  //for(int i = 2;i<4; i ++)  pinMode(i,INPUT);

  pinMode(encoder0pinB,INPUT);
  pinMode(encoder1pinB,INPUT);
  attachInterrupt(LF, LwheelSpeed, CHANGE);
  attachInterrupt(RT, RwheelSpeed, CHANGE);
}


/************************************************* Motor Control ***********************************************/

void Motor(int value,byte whichwheel)
{
  value = constrain(value,1000,2000);
  /*
  MotorLeft.write(90);
   MotorRight.write(90);
   */
  if(whichwheel == LF){
    MotorLeft.writeMicroseconds(value);
//    MotorLeft.write(100);
  }
  else if(whichwheel == RT){
    MotorRight.writeMicroseconds(value);
//    MotorRight.write(100);
  }
}


/************************************************ calculate speed (cm/s) ***********************************************/

float lastspeed(int longs,int diff)
{
  double internum = _perimeterA;
  internum /= diff;
  float dist = float(internum*longs);
  dist /= _FirmPulsePG;
  return dist;
}

void ResentSpeed()
{
  static int pasttime;
  static unsigned long lasttime;
  static unsigned long now;
  now = millis();
  pasttime = now - lasttime;
  lasttime = now;

  _speedleft = lastspeed(Lduration , pasttime);
  _speedright = lastspeed(Rduration , pasttime);

  //  Serial.println(pasttime);
  //  Serial.print("  ");
  //  Serial.println(Lduration);
  
  pastCoder[LF] += Lduration;
  pastCoder[RT] += Rduration;

  Lduration = 0;
  Rduration = 0;
}

/************************************************* PID Control ***********************************************/


float TVPIDcal(float prevspeed,boolean target)
{
  static int sumerror[2];
  static int i;
  if(target)
    i = 0;
  else
    i = 1;
  int derror;
  int error = _speedtarget[i] - prevspeed;

  sumerror[i] += error;
  sumerror[i] = min(_maximum,sumerror[i]);//limit the range of intergral segment
  sumerror[i] = max(_minimum,sumerror[i]);

  derror = _lasterror[i] - _preverror[i];
  _preverror[i] = _lasterror[i];
  _lasterror[i] = error;

  return (_proportion*error+_integral*sumerror[i]+_derivative*derror);
}

int TVAffect(float pidpara)
{
  float result = 500;
  float factor;

  if(pidpara>_maximum)
    factor = 1;
  else if(pidpara>0)
    factor = pidpara/_maximum;
  else if(pidpara<_minimum)
    factor = -1;
  else
    factor = pidpara/_maximum;

  result *= factor;
  result += 1500;

  return result;
}

/************************** Interrupt *****************************/


void LwheelSpeed()
{
  int Lstate = digitalRead(encoder0pinA);
  if((encoder0PinALast == LOW)&&Lstate==HIGH)
  {
    int val = digitalRead(encoder0pinB);
    if(val == LOW && LcoderDir)  LcoderDir = false; //Lreverse
    else if(val == HIGH && !LcoderDir)  LcoderDir = true;  //Lforward
  }
  encoder0PinALast = Lstate;

  if(!LcoderDir)  Lduration++;
  else  Lduration--;

  if(true)  RotationCoder[LF] ++;
  else RotationCoder[LF] = 0;
}


void RwheelSpeed()
{
  int Rstate = digitalRead(encoder1pinA);
  if((encoder1PinALast == LOW)&&Rstate==HIGH)
  {
    int val = digitalRead(encoder1pinB);
    if(val == LOW && RcoderDir)  RcoderDir = false; //Rreverse
    else if(val == HIGH && !RcoderDir)  RcoderDir = true;  //Rforward
  }
  encoder1PinALast = Rstate;

  if(!RcoderDir)  Rduration++;
  else  Rduration--;

  if(true)  RotationCoder[RT] ++;
  else RotationCoder[RT] = 0;
}

/********************** I2C Slaver ***********************

byte ReceiveByte[10];
unsigned long timer = 0;

void receiveEvent(int HowMany)
{
  unsigned long St = micros();
  int length = 0;
//  Serial.println(micros());

  if(Wire.available())
  {

    byte HeaderByte = Wire.receive();
    if(HeaderByte == 0x46)
    {
      int i;
      //delay(10);
      for(i = 0 ;i <10 ;i ++)  ReceiveByte[i] = 0;
      i = 0;
//      Serial.print("OK\t");
      ReceiveByte[i] = HeaderByte;
      i ++;
      while(Wire.available())
      {
        ReceiveByte[i] = Wire.receive();
        //Serial.print(ReceiveByte[i],BYTE);
        length = i;
        if(ReceiveByte[5] == 0x0A || (micros() - St > 2000))
        { 
          while(Wire.available())
          {
            byte BufferClear = Wire.receive();
          }
          break;
        }
        i ++;
      }
      //Serial.println();
      //Serial.println(micros() - St);
      //while(Wire.available())
    }
    else
    {
      Serial.print("Err");
      while(Wire.available())
      {
        byte BufferClear = Wire.receive();
      }
      //Serial.println(micros() - St);
    }
  }
  
//  for( int j = 0;j <= length;j++)
//  {
//    Serial.print(j);
//    Serial.print(":");
//    Serial.println(ReceiveByte[j],HEX);
//  }
//  Serial.println();
  
  
  if(ReceiveByte[5] == 0x0A)  
  {
//    Serial.print(ReceiveByte[2],HEX);
//     Serial.print(",");
//     Serial.print(ReceiveByte[3],HEX);
//     Serial.print(",");
//     Serial.print(ReceiveByte[4],HEX);
//     Serial.println("");
    wheelDir = ReceiveByte[2];
    _CMDspeed[LF] = ReceiveByte[3];
    _CMDspeed[RT] = ReceiveByte[4];

    if(wheelDir == 0x00||wheelDir == 0x01)
      _speedtarget[LF] = _CMDspeed[LF];
    else
      _speedtarget[LF] = _CMDspeed[LF]*(-1);

    if(wheelDir == 0x00||wheelDir == 0x10)
      _speedtarget[RT] = _CMDspeed[RT];
    else
      _speedtarget[RT] = _CMDspeed[RT]*(-1);

    for(int i = 0; i < 2; i++)  _speedtarget[i] = constrain(_speedtarget[i],-40,40);
//     Serial.print(millis() - timer);
//     timer = millis();
     
//     Serial.print(_speedtarget[LF]);
//     Serial.print(",");
//     Serial.print(_speedtarget[RT]);
//     Serial.println("");
     
  }
}

void requestEvent()
{  
  byte RequestString[Pbn] = {
    'F','E',0,0,0,0,0,10  };
  
  RequestString[2] = 0x00;
  
  if(pastCoder[LF] < 0)  bitWrite(RequestString[2],4,1);
  if(pastCoder[RT] < 0)  bitWrite(RequestString[2],0,1);
//  Serial.println(RequestString[2],HEX);
  
  RequestString[3] = abs(pastCoder[LF]) / 256;  
  RequestString[4] = abs(pastCoder[LF]) % 256; 
  RequestString[5] = abs(pastCoder[RT]) / 256;  
  RequestString[6] = abs(pastCoder[RT]) % 256; 
  
  totalCoder[LF] += pastCoder[LF];
  totalCoder[RT] += pastCoder[RT];
  
  pastCoder[LF] = 0;
  pastCoder[RT] = 0;
  
  Wire.send(RequestString,Pbn);
  
//  for( int j = 0;j < Pbn;j++)
//  {
//    Serial.print(j);
//    Serial.print(":");
//    Serial.println(RequestString[j],HEX);
//  }
//  Serial.println();
  
}
*/





